from pathlib import Path

import pytest


@pytest.fixture()
def samples_path():
    """Return the path to the samples."""
    return Path(__file__).parent / "samples"


def pytest_generate_tests(metafunc):
    """Generate test scenarios.

    See
    https://docs.pytest.org/en/7.1.x/example/parametrize.html#a-quick-port-of-testscenarios
    """
    id_list = []
    arg_values = []
    if metafunc.cls is None:
        return
    for scenario in metafunc.cls.scenarios:
        id_list.append(scenario[0])
        items = scenario[1].items()
        arg_names = [x[0] for x in items]
        arg_values.append([x[1] for x in items])
    metafunc.parametrize(arg_names, arg_values, ids=id_list, scope="class")



import multiprocessing
import warnings
from abc import ABC, abstractmethod
from contextlib import nullcontext, suppress
from pathlib import Path
from typing import Dict, Iterable, Iterator, Optional, Tuple, Union

import numpy as np
import xarray as xr
import zarr

from wsic.codecs import register_codecs
from wsic.enums import Codec, ColorSpace
from wsic.magic import summon_file_types
from wsic.metadata import ngff
from wsic.typedefs import PathLike
from wsic.utils import (
    TimeoutWarning,
    block_downsample_shape,
    main_process,
    mean_pool,
    mosaic_shape,
    ppu2mpp,
    resize_array,
    scale_to_fit,
    tile_slices,
)


class Reader(ABC):
    """Base class for readers."""

    def __init__(self, path: PathLike):
        """Initialize reader.

        Args:
            path (PathLike):
                Path to file.
        """
        self.path = Path(path)

    @abstractmethod
    def __getitem__(self, index: Tuple[Union[int, slice], ...]) -> np.ndarray:
        """Get pixel data at index."""
        raise NotImplementedError  # pragma: no cover

    @classmethod
    def from_file(cls, path: Path) -> "Reader":
        """Return reader for file.

        Args:
            path (Path): Path to file.

        Returns:
            Reader: Reader for file.
        """
        path = Path(path)
        file_types = summon_file_types(path)
        if ("jp2",) in file_types:
            return JP2Reader(path)
        with suppress(ImportError):
            import openslide

            with suppress(openslide.OpenSlideError):
                return OpenSlideReader(path)
        if ("tiff",) in file_types:
            return TIFFReader(path)
        if ("dicom",) in file_types or ("dcm",) in file_types:
            return DICOMWSIReader(path)
        if ("zarr",) in file_types:
            return ZarrReader(path)
        raise ValueError(f"Unsupported file type: {path}")

    def thumbnail(self, shape: Tuple[int, ...], approx_ok: bool = False) -> np.ndarray:
        """Generate a thumbnail image of (or near) the requested shape.

        Args:
            shape (Tuple[int, ...]):
                Shape of the thumbnail.
            approx_ok (bool):
                If True, return a thumbnail that is approximately the
                requested shape. It will be equal to or larger than the
                requested shape (the next largest shape possible via
                an integer block downsampling).

        Returns:
            np.ndarray: Thumbnail.
        """
        # NOTE: Assuming first two are Y and X
        yx_shape = self.shape[:2]
        yx_tile_shape = (
            self.tile_shape[:2] if self.tile_shape else np.minimum(yx_shape, (256, 256))
        )
        self_mosaic_shape = self.mosaic_shape or mosaic_shape(yx_shape, yx_tile_shape)
        (
            downsample_shape,
            downsample_tile_shape,
            downsample,
        ) = self._find_thumbnail_downsample(shape, yx_shape, yx_tile_shape)
        # NOTE: Assuming channels last
        channels = self.shape[-1]
        thumbnail = np.zeros(downsample_shape + (channels,), dtype=np.uint8)
        # Resize tiles to new_downsample_tile_shape and combine
        tile_indexes = list(np.ndindex(self_mosaic_shape))
        for tile_index in self.pbar(tile_indexes, desc="Generating thumbnail"):
            try:
                tile = self.get_tile(tile_index)
            except (ValueError, NotImplementedError):  # e.g. Not tiled
                tile = self[tile_slices(tile_index, yx_tile_shape)]
            tile = mean_pool(tile.astype(float), downsample).astype(np.uint8)
            # Make sure the tile being written will not exceed the
            # bounds of the thumbnail
            yx_position = tuple(
                i * size for i, size in zip(tile_index, downsample_tile_shape)
            )
            max_y, max_x = (
                min(tile_max, thumb_max - position)
                for tile_max, thumb_max, position in zip(
                    tile.shape, thumbnail.shape, yx_position
                )
            )
            sub_tile = tile[:max_y, :max_x]
            thumbnail[tile_slices(tile_index, downsample_tile_shape)] = sub_tile
        return thumbnail if approx_ok else resize_array(thumbnail, shape, "bicubic")

    @staticmethod
    def _find_thumbnail_downsample(
        thumbnail_shape: Tuple[int, int],
        yx_shape: Tuple[int, int],
        yx_tile_shape: Tuple[int, int],
    ) -> Tuple[Tuple[int, int], Tuple[int, int], int]:
        """Find the downsample and tile shape for a thumbnail.

        Args:
            thumbnail_shape (Tuple[int, int]):
                Shape of the thumbnail to be generated.
            yx_shape (Tuple[int, int]):
                Shape of the image in Y and X.
            yx_tile_shape (Tuple[int, int]):
                Shape of the tiles in Y and X which will be used to
                generate the thumbnail.

        Returns:
            Tuple[Tuple[int, int], Tuple[int, int], int]:
                Shape of the downsampled image, shape of the downsampled
                tiles, and the downsample factor.
        """
        downsample_shape = yx_shape
        downsample_tile_shape = yx_tile_shape
        downsample = 0
        while True:
            new_downsample = downsample + 1
            next_downsample_shape, new_downsample_tile_shape = block_downsample_shape(
                yx_shape, new_downsample, yx_tile_shape
            )
            if all(x <= max(0, y) for x, y in zip(downsample_shape, thumbnail_shape)):
                break
            downsample_shape = next_downsample_shape
            downsample_tile_shape = new_downsample_tile_shape
            downsample = new_downsample
        return downsample_shape, downsample_tile_shape, downsample

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        # Naive base implementation using __getitem__
        if not decode:
            raise NotImplementedError(
                "Fetching tiles without decoding is not supported."
            )
        if not hasattr(self, "tile_shape"):
            raise ValueError(
                "Cannot get tile from a non-tiled reader"
                " (must have attr 'tile_shape')."
            )
        slices = tile_slices(index, self.tile_shape)
        return self[slices]

    @staticmethod
    def pbar(iterable: Iterable, *args, **kwargs) -> Iterator:
        """Return an iterator that displays a progress bar.

        Uses tqdm if installed, otherwise falls back to a simple iterator.

        Args:
            iterable (Iterable):
                Iterable to iterate over.
            args (tuple):
                Positional arguments to pass to tqdm.
            kwargs (dict):
                Keyword arguments to pass to tqdm.

        Returns:
            Iterator: Iterator that displays a progress bar.
        """
        try:
            from tqdm.auto import tqdm
        except ImportError:

            def tqdm(x, *args, **kwargs):
                return x

        return tqdm(iterable, *args, **kwargs)

    @property
    def original_shape(self) -> Tuple[int, ...]:
        """Return the original shape of the image."""
        return self.shape


class JP2Reader(Reader):
    """Reader for JP2 files using glymur.

    Args:
        path (Path): Path to file.
    """

    def __init__(self, path: Path) -> None:
        super().__init__(path)
        import glymur

        # Enable multithreading
        if glymur.options.version.openjpeg_version_tuple >= (2, 2, 0):
            glymur.set_option("lib.num_threads", multiprocessing.cpu_count())

        self.jp2 = glymur.Jp2k(str(path))
        self.shape = self.jp2.shape
        self.dtype = np.uint8
        self.axes = "YXS"
        self.microns_per_pixel = self._get_mpp()
        self.tile_shape = self._get_tile_shape()

    def _get_mpp(self) -> Optional[Tuple[float, float]]:
        """Get the microns per pixel for the image.

        Returns:
            Optional[Tuple[float, float]]:
                The resolution of the image in microns per pixel.
                If the resolution is not available, this will be None.
        """
        import glymur

        boxes = {type(box): box for box in self.jp2.box}
        if not boxes:
            warnings.warn(
                "Cannot get MPP. No boxes found, invalid JP2 file.", stacklevel=2
            )
            return None
        header_box = boxes[glymur.jp2box.JP2HeaderBox]
        header_sub_boxes = {type(box): box for box in header_box.box}
        resolution_box = header_sub_boxes.get(glymur.jp2box.ResolutionBox)
        if resolution_box is None:
            return None
        resolution_sub_boxes = {type(box): box for box in resolution_box.box}
        capture_resolution_box = resolution_sub_boxes.get(
            glymur.jp2box.CaptureResolutionBox
        )
        if capture_resolution_box is None:
            return None
        # Read the resolution capture box in grid points (pixels) / meter
        pixels_per_meter_y = capture_resolution_box.vertical_resolution
        pixels_per_meter_x = capture_resolution_box.horizontal_resolution
        return ppu2mpp(pixels_per_meter_x, "m"), ppu2mpp(pixels_per_meter_y, "m")

    def _get_tile_shape(self) -> Tuple[int, int]:
        """Get the tile shape as a (height, width) tuple.

        Returns:
            Tuple[int, int]:
                The tile shape as (height, width).
        """
        import glymur

        boxes = {type(box): box for box in self.jp2.box}
        ccb = boxes.get(glymur.jp2box.ContiguousCodestreamBox)
        if ccb is None:
            raise ValueError("No codestream box found.")
        segments = {type(segment): segment for segment in ccb.codestream.segment}
        siz = segments.get(glymur.codestream.SIZsegment)
        if siz is None:
            raise ValueError("No SIZ segment found.")
        return (siz.ytsiz, siz.xtsiz)

    def __getitem__(self, index: tuple) -> np.ndarray:
        """Get pixel data at index."""
        return self.jp2[index]

    def thumbnail(self, shape: Tuple[int, ...], approx_ok: bool = False) -> np.ndarray:
        scale = scale_to_fit(self.shape[:2], shape)
        downsample = 1 / scale
        out_shape = tuple(int(x * scale) for x in self.shape[:2])
        # Glymur requires a power of two stride
        pow_2_downsample = 2 ** np.floor(np.log2(downsample))
        # Get the power of two downsample
        thumbnail = self.jp2[::pow_2_downsample, ::pow_2_downsample]
        # Resize the thumbnail if required
        if approx_ok:
            return thumbnail
        return resize_array(thumbnail, out_shape)


class TIFFReader(Reader):
    """Reader for TIFF files using tifffile."""

    def __init__(self, path: Path) -> None:
        """Initialize reader.

        Args:
            path (Path): Path to file.
        """
        import tifffile

        super().__init__(path)
        self._tiff = tifffile.TiffFile(str(path))
        self._tiff_page = self._tiff.pages[0]
        self.microns_per_pixel = self._get_mpp()

        # Handle reading as an xarray dataset
        self._zarr_tiff_store = self._tiff.aszarr()
        # Ensure that the zarr store is a group
        if zarr.storage.contains_array(self._zarr_tiff_store):
            # Create an in memory group
            group = zarr.group()
            # Add the array to the group under the key "0"
            array = zarr.open(self._zarr_tiff_store, mode="r")
            group[0] = array
            # Copy attrs over to the group member (a zarr bug?)
            if "_ARRAY_DIMENSIONS" not in array.attrs:
                raise ValueError(
                    f"No _ARRAY_DIMENSIONS found in {self._zarr_tiff_store}."
                )
            for key, value in array.attrs.items():
                group[0].attrs[key] = value
            self._zarr_tiff_store = group.store
        # Open the store as an xarray dataset
        self._zarr = zarr.open(self._zarr_tiff_store, mode="r")
        self._dataset = xr.open_zarr(self._zarr_tiff_store, consolidated=False)
        # Copy over the dtype of the dataset (xarray bug?)
        for key, array in self._zarr.items():
            self._dataset[key] = self._dataset[key].astype(array.dtype)
        # Rename S to C
        if "S" in self._dataset["0"].dims:
            self._dataset["0"] = self._dataset["0"].rename({"S": "C"})
        # Normalise axes to TZYXC
        self._tzyxc_dataset = self._dataset.copy()
        self._tzyxc_dataset["0"] = self._tzyxc_dataset["0"].expand_dims(
            dim=[a for a in "TCZYX" if a not in self._tzyxc_dataset["0"].dims],
        )
        self._tzyxc_dataset["0"] = self._tzyxc_dataset["0"].transpose(
            "T", "Z", "Y", "X", "C"
        )

        # Set default time and depth
        self.default_t = 0
        self.default_z = 0

        # Set standard Reader attributes
        self.shape = self._dataset["0"].shape
        self.dtype = self._dataset["0"].dtype
        self.is_tiled = self._tiff_page.is_tiled
        self.tile_shape = None
        self.mosaic_shape = None
        self.mosaic_byte_offsets = None
        self.mosaic_byte_counts = None
        # Read tile shape if tiled
        if self.is_tiled:
            self.tile_shape = (self._tiff_page.tilelength, self._tiff_page.tilewidth)
            self.mosaic_shape = mosaic_shape(
                array_shape=self._tiff_page.shape, tile_shape=self.tile_shape
            )
            self.mosaic_byte_offsets = np.array(self._tiff_page.dataoffsets).reshape(
                self.mosaic_shape
            )
            self.mosaic_byte_counts = np.array(self._tiff_page.databytecounts).reshape(
                self.mosaic_shape
            )
        self.jpeg_tables = self._tiff_page.jpegtables
        self.color_space: ColorSpace = ColorSpace.from_tiff(self._tiff_page.photometric)
        self.codec: Codec = Codec.from_tiff(self._tiff_page.compression)
        self.compression_level = None  # To be filled in if known later

    @property
    def original_shape(self) -> Tuple[int, ...]:
        return self._tiff_page.shape

    def _get_mpp(self) -> Optional[Tuple[float, float]]:
        """Get the microns per pixel for the image.

        This checks the resolution and resolution unit TIFF tags.

        Returns:
            Optional[Tuple[float, float]]:
                The resolution of the image in microns per pixel.
                If the resolution is not available, this will be None.
        """
        if self._tiff.is_svs:
            return self._get_mpp_svs()
        try:
            return self._get_mpp_tiff_res_tag()
        except KeyError:
            return None

    def _get_mpp_tiff_res_tag(self):
        """Get the microns per pixel for the image from the TIFF tags."""
        tags = self._tiff_page.tags
        y_resolution = tags["YResolution"].value[0] / tags["YResolution"].value[1]
        x_resolution = tags["XResolution"].value[0] / tags["XResolution"].value[1]
        resolution_units = tags["ResolutionUnit"].value
        return ppu2mpp(x_resolution, resolution_units), ppu2mpp(
            y_resolution, resolution_units
        )

    @staticmethod
    def _parse_svs_key_values(description: str) -> Dict[str, str]:
        """Parse the key value pairs from the SVS description."""
        parts = description.split("\n")
        # Header in parts[0]
        key_values_str = parts[-1]
        # Conver to dict
        return {
            kv.split("=")[0].strip(): kv.split("=")[1].strip()
            for kv in key_values_str.split("|")
        }

    def _get_mpp_svs(self):
        """Get the microns per pixel for the image from the SVS description."""
        if self._tiff_page.description is None:
            return None
        svs_key_values = self._parse_svs_key_values(self._tiff_page.description)
        mpp = svs_key_values.get("MPP")
        return (float(mpp), float(mpp)) if mpp else None

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        if self.tile_shape is None:
            raise ValueError("Image is not tiled.")
        flat_index = index[0] * self.tile_shape[1] + index[1]
        fh = self._tiff.filehandle
        _ = fh.seek(self.mosaic_byte_offsets[index])
        data = fh.read(self.mosaic_byte_counts[index])
        if not decode:
            return data
        tile, _, shape = self._tiff_page.decode(
            data, flat_index, jpegtables=self._tiff_page.jpegtables
        )
        # tile may be None e.g. with NDPI blank tiles
        if tile is None:
            tile = np.zeros(shape, dtype=self.dtype)
        return tile[0]

    def __getitem__(self, index: Tuple[Union[slice, int]]) -> np.ndarray:
        """Get pixel data at index."""
        index = index if isinstance(index, tuple) else (index,)
        index = (self.default_t, self.default_z) + index
        return self._tzyxc_dataset["0"][index].as_numpy().data

    def thumbnail(self, shape: Tuple[int, ...], approx_ok: bool = False) -> np.ndarray:
        """Get a thumbnail of the image.

        Uses xarray/dask block map to get the thumbnail.

        Args:
            shape (Tuple[int, ...]):
                The shape of the thumbnail to get.
            approx_ok (bool, optional):
                Whether to use an approximate thumbnail. Defaults to False.

        Returns:
            np.ndarray:
                The thumbnail of the image.
        """
        yxc = self._tzyxc_dataset["0"][self.default_t, self.default_z]
        (
            _,
            _,
            downsample,
        ) = self._find_thumbnail_downsample(
            shape,
            yxc.shape[:2],
            yxc.data.chunksize[:2],
        )
        thumbnail = (  # noqa: ECE001
            yxc.coarsen(X=downsample, boundary="trim")
            .mean()
            .coarsen(Y=downsample, boundary="trim")
            .mean()
            .astype("u1")
            .as_numpy()
        )
        return (
            thumbnail.data
            if approx_ok
            else resize_array(thumbnail.data, shape, "bicubic")
        )


class DICOMWSIReader(Reader):
    """Reader for DICOM Whole Slide Images (WSIs) using wsidicom.

    DICOM Whole Slide Imaging:  https://dicom.nema.org/Dicom/DICOMWSI/
    """

    def __init__(self, path: Path) -> None:
        """Initialize reader.

        Args:
            path (Path):
                Path to file.
        """
        from pydicom import Dataset
        from wsidicom import WsiDicom

        super().__init__(path)

        # Set up a timeout warning for slow sparse tiled files
        sparse_tiled_warning = TimeoutWarning(
            "Looks like this is taking a while..."
            "if your DICOM file has a 'DimensionOrganizationType' "
            "of 'TILED_SPARSE',  this may be causing it to be slow.",
            timeout=1,
        )
        context = sparse_tiled_warning if main_process() else nullcontext()

        with context:
            # Open the file, this will take a while if the file is sparse tiled
            self.slide = WsiDicom.open(self.path)

        channels = len(self.slide.read_tile(0, (0, 0)).getbands())
        self.shape = (self.slide.size.height, self.slide.size.width, channels)
        self.dtype = np.uint8
        self.microns_per_pixel = (
            self.slide.levels.base_level.mpp.height,
            self.slide.levels.base_level.mpp.width,
        )
        dataset: Dataset = self.slide.levels.base_level.datasets[0]
        self.tile_shape = (dataset.Rows, dataset.Columns)
        self.mosaic_shape = mosaic_shape(
            self.shape,
            self.tile_shape,
        )
        # Sanity check
        if np.prod(self.mosaic_shape[:2]) != int(dataset.NumberOfFrames):
            raise ValueError(
                f"Number of frames in DICOM dataset {dataset.NumberOfFrames}"
                f" does not match mosaic shape {self.mosaic_shape}."
            )
        self.codec = Codec.NONE
        if hasattr(dataset, "LossyImageCompressionMethod"):
            self.codec: Codec = Codec.from_string(dataset.LossyImageCompressionMethod)
        self.compression_level = (
            None  # Set if known: dataset.get(LossyImageCompressionRatio)?
        )
        self.color_space = ColorSpace.from_dicom(dataset.photometric_interpretation)
        self.jpeg_tables = None

    def performance_check(self) -> None:
        """Check attributes of the file and warn if they are not optimal.

        A 'DimensionOrganizationType' of `TILED_SPARSE` versus
        `TILED_FULL` will negatively impact performance.

        """
        from pydicom import Dataset

        dataset: Dataset = self.slide.levels.base_level.datasets[0]
        if dataset.DimensionOrganizationType != "TILED_FULL":
            warnings.warn(
                "DICOM file is not TILED_FULL. Performance may be impacted."
                " Consider converting to TILED_FULL like so:\n"
                "\n>>> from wsidicom import WsiDicom"
                "\n>>> with WsiDicom.open(path_to_input) as slide:"
                "\n>>>     slide.save(path_to_ouput)"
                "\nThis is lossless and fast.",
                stacklevel=2,
            )
            should_convert = input(
                "Would you like to create a TILED_FULL copy now? [y/n]"
            )
            if should_convert.lower().strip() == "y":
                self._make_full_tiled_copy()

    def _make_full_tiled_copy(self) -> None:
        """Make a copy of the file with TILED_FULL DimensionOrganizationType."""
        print("Converting to TILED_FULL...")
        from wsidicom import WsiDicom

        if self.path.is_dir():
            new_path = self.path.with_suffix(".tiled_full")
            new_path.mkdir(parents=True, exist_ok=False)
        else:
            new_path = self.path.with_suffix(".tiled_full.dcm")
        self.slide.save(new_path)
        self.path = new_path
        self.slide = WsiDicom.open(self.path)
        print("Done.")

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        if decode:
            return np.array(self.slide.read_tile(level=0, tile=index[::-1], z=0))
        return self.slide.read_encoded_tile(level=0, tile=index[::-1], z=0)

    def __getitem__(self, index: Tuple[Union[slice, int]]) -> np.ndarray:
        """Get pixel data at index."""
        if index is ...:
            return np.array(self.slide.levels.base_level.get_default_full())
        xs = index[1]
        ys = index[0]
        start_x = xs.start or 0
        start_y = ys.start or 0
        end_x = xs.stop or self.shape[1]
        end_y = ys.stop or self.shape[0]

        # Prevent reading past the edges of the image
        end_x = min(end_x, self.shape[1])
        end_y = min(end_y, self.shape[0])

        # Read the image
        img = self.slide.read_region(
            location=(start_x, start_y),
            level=0,
            size=(end_x - start_x, end_y - start_y),
            z=0,
        )
        return np.array(img.convert("RGB"))


class OpenSlideReader(Reader):
    """Reader for OpenSlide files using openslide-python."""

    def __init__(self, path: Path) -> None:
        import openslide

        super().__init__(path)
        self.os_slide = openslide.OpenSlide(str(path))
        self.shape = self.os_slide.level_dimensions[0][::-1] + (3,)
        self.dtype = np.uint8
        self.axes = "YXS"
        self.tile_shape = None  # No easy way to get tile shape currently
        self.microns_per_pixel = self._get_mpp()
        self.mosaic_shape = None

    def get_tile(self, index: Tuple[int, int], decode: bool = True) -> np.ndarray:
        """Get tile at index.

        Args:
            index (Tuple[int, int]):
                The index of the tile to get.
            decode (bool, optional):
                Whether to decode the tile. Defaults to True.

        Returns:
            np.ndarray:
                The tile at index.
        """
        raise NotImplementedError("OpenSlideReader does not support reading tiles.")

    def _get_mpp(self) -> Optional[Tuple[float, float]]:
        """Get the microns per pixel for the image.

        Returns:
            Optional[Tuple[float, float]]:
                The microns per pixel as (x, y) tuple.
        """
        try:
            return (
                float(self.os_slide.properties["openslide.mpp-x"]),
                float(self.os_slide.properties["openslide.mpp-y"]),
            )
        except KeyError:
            warnings.warn("OpenSlide could not find MPP.", stacklevel=2)
        # Fall back to TIFF resolution tags
        try:
            resolution = (
                float(self.os_slide.properties["tiff.XResolution"]),
                float(self.os_slide.properties["tiff.YResolution"]),
            )
            units = self.os_slide.properties["tiff.ResolutionUnit"]
            self._check_sensible_resolution(resolution, units)
            return tuple(ppu2mpp(x, units) for x in resolution)
        except KeyError:
            warnings.warn("No resolution metadata found.", stacklevel=2)
        return None

    @staticmethod
    def _check_sensible_resolution(
        tiff_resolution: Tuple[float, float], tiff_units: int
    ) -> None:
        """Check whether the resolution is sensible.

        It is common for TIFF files to have incorrect resolution tags.
        This method checks whether the resolution is sensible and warns
        if it is not.

        Args:
            tiff_resolution (Tuple[float, float]):
                The TIFF resolution as an (x, y) tuple.
            tiff_units (int):
                The TIFF units of the resolution. A value of 2 indicates
                inches and a value of 3 indicates centimeters.
        """
        if tiff_units == 2 and 72 in tiff_resolution:
            warnings.warn(
                "TIFF resolution tags found."
                " However, they have a common default value of 72 pixels per inch."
                " This may from a misconfigured software library or tool"
                " which is expecting to handle print documents.",
                stacklevel=2,
            )
        if 0 in tiff_resolution:
            warnings.warn(
                "TIFF resolution tags found."
                " However, one or more of the values is zero.",
                stacklevel=2,
            )

    def __getitem__(self, index: Tuple[Union[int, slice], ...]) -> np.ndarray:
        """Get pixel data at index."""
        if index is ...:
            return np.array(self.os_slide.get_thumbnail(self.os_slide.dimensions))
        xs = index[1]
        ys = index[0]
        start_x = xs.start or 0
        start_y = ys.start or 0
        end_x = xs.stop or self.shape[1]
        end_y = ys.stop or self.shape[0]

        # Prevent reading past the edges of the image
        end_x = min(end_x, self.shape[1])
        end_y = min(end_y, self.shape[0])

        # Read the image
        img = self.os_slide.read_region(
            location=(start_x, start_y),
            level=0,
            size=(end_x - start_x, end_y - start_y),
        )
        return np.array(img.convert("RGB"))


class ZarrReader(Reader):
    """Reader for zarr files."""

    def __init__(self, path: PathLike, axes: Optional[str] = None) -> None:
        super().__init__(path)
        register_codecs()
        self.zarr = zarr.open(str(path), mode="r")
        # Currently mpp not stored in zarr, could use xarray metadata
        # for this or custom wsic metadata
        self.microns_per_pixel = None

        # If it is a zarr array, put it in a group
        if isinstance(self.zarr, zarr.Array):
            group = zarr.group()
            group[0] = self.zarr
            self.zarr = group

        self.shape = self.zarr[0].shape
        if len(self.shape) not in (2, 3, 4, 5):
            raise ValueError(
                "Only Zarrs with between 2 (e.g. YX) and 5 (e.g. TCZYX) "
                "dimensions are supported."
            )
        self.dtype = self.zarr[0].dtype

        # Assume the axes, this will be used if not given or specified in metadata
        assumed_axes_mapping = {
            2: "YX",
            3: "YXC",
            4: "CZYX",
            5: "TCZYX",
        }
        self.axes = assumed_axes_mapping[len(self.shape)]

        # Use the NGFF metadata NGFF if present
        self.is_ngff = "omero" in self.zarr.attrs
        self.zattrs = None
        if self.is_ngff:
            self.zattrs = self._load_zattrs()
            self.axes = "".join(
                multiscale.axis.name for multiscale in self.zattrs.multiscales
            ).upper()
        # Use the given axes if not None
        self.axes = axes or self.axes

        # Tile shape and mosaic attrs
        self.tile_shape = self.zarr[0].chunks[:2]
        self.mosaic_shape = mosaic_shape(self.shape, self.tile_shape)

    def __getitem__(self, index: Tuple[Union[int, slice], ...]) -> np.ndarray:
        return self.zarr[0][index]

    def _load_zattrs(self) -> ngff.Zattrs:
        """Load the zarr attrs dictionary into dataclasses."""
        return ngff.Zattrs(
            _creator=ngff.Creator(**self.zattrs.get("_creator")),
            multiscales=[
                ngff.Multiscale(
                    axes=[ngff.Axis(**axis) for axis in multiscale.get("axes", [])],
                    datasets=[
                        ngff.Dataset(
                            path=dataset.get("path"),
                            coordinateTransformations=[
                                ngff.CoordinateTransformation(
                                    **coordinate_transformation
                                )
                                for coordinate_transformation in dataset.get(
                                    "coordinateTransformations", []
                                )
                            ],
                        )
                        for dataset in multiscale.get("datasets", [])
                    ],
                    version=multiscale.get("version"),
                )
                for multiscale in self.zarr.attrs.get("multiscales", [])
            ],
            _ARRAY_DIMENSIONS=self.zarr.attrs.get("_ARRAY_DIMENSIONS"),
            omero=ngff.Omero(
                name=self.zarr.attrs.get("omero", {}).get("name"),
                channels=[
                    ngff.Channel(
                        name=channel.get("name"),
                        coefficient=channel.get("coefficient"),
                        color=channel.get("color"),
                        family=channel.get("family"),
                        inverted=channel.get("inverted"),
                        label=channel.get("label"),
                        window=ngff.Window(**channel.get("window", {})),
                    )
                    for channel in self.zattrs.get("omero", {}).get("channels", [])
                ],
            ),
        )



import copy
import warnings

import pytest

import wsic


@pytest.fixture()
def _register_codecs():
    """Register codecs for the duration of the test function."""
    # Register codecs
    import numcodecs

    from wsic.codecs import register_codecs

    codec_registry_backup = copy.deepcopy(numcodecs.registry.codec_registry)
    register_codecs()

    # Yield to the test function
    yield

    # Teardown
    numcodecs.registry.codec_registry = codec_registry_backup


def test_condensed():
    """Test that the condensed string contains not spaces."""
    for value in wsic.enums.Codec:
        condensed = value.condensed()
        assert " " not in condensed


def test_from_string_fail():
    with pytest.raises(ValueError, match="Unknown codec"):
        wsic.enums.Codec.from_string("unknown")


def test_from_tiff_fail():
    with pytest.raises(ValueError, match="Unknown TIFF compression"):
        wsic.enums.Codec.from_tiff(1234)


@pytest.mark.usefixtures("_register_codecs")
@pytest.mark.parametrize("codec", wsic.enums.Codec)
def test_to_numcodecs_config(codec):
    import numcodecs

    # Skip unsupported codecs
    if codec in (wsic.enums.Codec.LZ77,):
        return
    try:
        config = codec.to_numcodecs_config()
    except ValueError:
        warnings.warn(f"{codec} is not supported.", stacklevel=2)
        return

    # Skip if NONE
    if codec is wsic.enums.Codec.NONE:
        return
    assert "id" in config

    # JPEG200 special cases
    if codec in (wsic.enums.Codec.JPEG2000, wsic.enums.Codec.J2K):
        assert "codecformat" in config
    numcodecs.get_codec(config)


def test_to_numcodecs_config_fail():
    with pytest.raises(ValueError, match="not a supported codec"):
        wsic.enums.Codec.LZ77.to_numcodecs_config()



"""Custom codecs for wsic."""

import warnings


def register_codecs() -> None:
    """Add additional codecs to the numcodecs registry.

    Codecs from imagecodecs include:
    - JPEG
    - PNG
    - DEFLATE
    - JPEG 2000
    - JPEG-LS
    - JPEG XR
    - JPEG XL
    - WebP

    """
    import numcodecs

    try:
        from imagecodecs.numcodecs import register_codecs as register_imagecodecs_codecs

        # Register imagecodecs codecs
        if "imagecodecs_jpeg" not in numcodecs.registry.codec_registry:
            register_imagecodecs_codecs()
    except ImportError:
        warnings.warn(
            "imagecodecs is not installed, some codecs will not be available.",
            stacklevel=2,
        )



import importlib
import struct
from datetime import datetime
from io import SEEK_END, SEEK_SET, BytesIO
from math import ceil
from pathlib import Path
from typing import Iterable, List, Literal, Optional, Tuple, Union

from pydicom.dataset import Dataset, FileMetaDataset
from pydicom.uid import JPEGBaseline, VLWholeSlideMicroscopyImageStorage, generate_uid

PathLike = Union[Path, str]
FileLike = Union[PathLike, BytesIO]


class CodeRef(Dataset):
    """Create a code reference dataset.

    These code datasets are used to refer to codified values in standards
    such as DICOM and SNOMED CT.
    """

    def __init__(
        self,
        value: str,
        designator: str,
        meaning: str,
    ) -> None:
        super().__init__()
        self.CodeValue = value
        self.CodingSchemeDesignator = designator
        self.CodeMeaning = meaning


def append_frames(
    io: FileLike,
    frame_iterable: Iterable[bytes],
    frame_count: int,
) -> None:
    """Append frames to an existing dataset."""
    sequence_delimitation_item_tag = (0xFFFE, 0xE0DD)
    item_tag = (0xFFFE, 0xE000)
    pixel_data_tag = (0x7FE0, 0x0010)
    other_binary_type = b"OB"
    undefined_length = 0xFFFFFFFF
    tag_struct = struct.Struct("<HH")
    length_struct = struct.Struct("<I")

    offsets = [0]
    lengths = []

    ds = Dataset()
    ds.is_little_endian = True
    ds.is_implicit_VR = False
    ds.ExtendedOffsetTable = struct.pack("<Q", 0) * frame_count
    ds.ExtendedOffsetTableLengths = struct.pack("<Q", 0) * frame_count
    table_bio = BytesIO()
    ds.save_as(table_bio, write_like_original=True)
    table_bytes = table_bio.getvalue()
    eot_offset = None

    with io if isinstance(io, BytesIO) else open(io, "rb+") as file:  # noqa: SIM115
        # Seek to the end of the file
        file.seek(0, SEEK_END)
        eot_offset = file.tell()

        # Write the blank extended offset table
        file.write(table_bytes)

        # Write the pixel data tag
        file.write(
            tag_struct.pack(*pixel_data_tag)  # Tag
            + other_binary_type  # VR
            + b"\x00\x00"  # Reserved
        )
        file.write(length_struct.pack(undefined_length))
        pixel_data_length = 0
        # Start with an empty basic offset table
        file.write(tag_struct.pack(*item_tag) + length_struct.pack(0))
        pixel_data_length += tag_struct.size + length_struct.size
        for n in range(frame_count):
            try:
                frame_bytes = next(frame_iterable)
            except StopIteration as error:
                raise ValueError(
                    f"Expected {frame_count} frames,"
                    f" but StopIteration raised after {n+1}."
                ) from error
            if len(frame_bytes) % 2:
                frame_bytes += b"\x00"
            pixel_data_element = (
                tag_struct.pack(*item_tag)
                + length_struct.pack(len(frame_bytes))
                + frame_bytes
            )
            pixel_data_length += len(pixel_data_element)
            file.write(pixel_data_element)
            lengths.append(len(frame_bytes))

        # Fill offsets list
        offsets.extend(
            offsets[i] + length + tag_struct.size + length_struct.size
            for i, length in enumerate(lengths[:-1])
        )

        # Sanity checks
        # Check number of offsets and lengths match frame count
        if not len(offsets) == len(lengths) == frame_count:
            raise ValueError(  # pragma: no cover
                f"Number of offsets ({len(offsets)}) and lengths ({len(lengths)}) "
                f"do not match frame count ({frame_count})."
            )
        # Check end of file is after the EOT offset
        if not file.tell() > eot_offset:
            raise ValueError(  # pragma: no cover
                f"End of file ({file.tell()}) is not after the EOT offset"
                f" ({eot_offset})."
            )

        # Write the extended offset table
        table_bio = BytesIO()
        ds.ExtendedOffsetTable = struct.pack(f"<{len(offsets)}Q", *offsets)
        ds.ExtendedOffsetTableLengths = struct.pack(f"<{len(lengths)}Q", *lengths)
        ds.save_as(table_bio, write_like_original=True)
        new_table_bytes = table_bio.getvalue()

        # Sanity check
        if len(new_table_bytes) != len(table_bytes):
            raise ValueError(  # pragma: no cover
                f"New extended offset table length ({len(new_table_bytes)}) "
                f"does not match original ({len(table_bytes)})."
            )
        # Write at the eot offset
        file.seek(eot_offset, SEEK_SET)
        file.write(new_table_bytes)

        # Write the sequence delimitation item
        file.seek(0, SEEK_END)
        file.write(
            tag_struct.pack(*sequence_delimitation_item_tag) + length_struct.pack(0)
        )


class PreparationStep(Dataset):
    """Abstraction for a Preparation Step Sequence item.

    Currently unused
    """

    def __init__(
        self,
        data_type: str,
        value: str,
        concept_name_sequence: List[Dataset],
        concept_code_sequence: Optional[List[Dataset]] = None,
    ) -> None:
        super().__init__()
        self.ValueType = data_type
        if data_type == "CODE":
            self.CodeValue = value
        elif data_type == "TEXT":
            self.TextValue = value
        else:
            raise ValueError("Unknown ValueType")
        self.ConceptNameCodeSequence = concept_name_sequence
        if concept_code_sequence is not None:
            self.ConceptCodeSequence = concept_code_sequence


def ffpe_he_preparation_sequence() -> List[Dataset]:
    """Create a Formalin Fixed Parafin Embedded H&E Stained Preparation Sequence."""
    specimen_id = Dataset()
    specimen_id.ValueType = "TEXT"  # Type 1
    specimen_id.TextValue = "D18-6003 A-1-1"  # Type 1
    substep_0_name_0 = Dataset()
    substep_0_name_0.CodeValue = "121041"  # "Specimen Identifier"
    substep_0_name_0.CodingSchemeDesignator = "DCM"  # DICOM defined code
    substep_0_name_0.CodeMeaning = "Specimen Identifier"
    specimen_id.ConceptNameCodeSequence = [substep_0_name_0]

    spec_id_issuer = Dataset()
    spec_id_issuer.ValueType = "TEXT"  # Type 1
    spec_id_issuer.TextValue = "XYZ Medical Centre"  # Type 1
    substep_1_name_0 = Dataset()
    substep_1_name_0.CodeValue = "111724"  # "Issuer of Specimen Identifier"
    substep_1_name_0.CodingSchemeDesignator = "DCM"  # DICOM defined code
    substep_1_name_0.CodeMeaning = "Issuer of Specimen Identifier"
    spec_id_issuer.ConceptNameCodeSequence = [substep_1_name_0]

    collection = Dataset()
    collection.ValueType = "CODE"  # Type 1
    collection_name = Dataset()
    collection_name.CodeValue = "111701"
    collection_name.CodingSchemeDesignator = "DCM"  # DICOM defined code
    collection_name.CodeMeaning = "Processing Type"
    collection_code = Dataset()
    collection_code.CodeValue = "17636008"  # 17636008/P3-02000 (CT/RT)
    collection_code.CodingSchemeDesignator = "SCT"  # Snowmed CT defined code
    collection_code.CodeMeaning = "Specimen Collection"
    collection.ConceptNameCodeSequence = [collection_name]
    collection.ConceptCodeSequence = [collection_code]

    taken = Dataset()
    taken.ValueType = "TEXT"
    taken.TextValue = "Taken"
    taken_name = Dataset()
    taken_name.CodeValue = "111703"
    taken_name.CodingSchemeDesignator = "DCM"
    taken_name.CodeMeaning = "Processing Step Description"
    taken.ConceptNameCodeSequence = [taken_name]

    excision = Dataset()
    excision.ValueType = "CODE"
    excision_name = Dataset()
    excision_name.CodeValue = "111704"
    excision_name.CodingSchemeDesignator = "DCM"
    excision_name.CodeMeaning = "Sampling Method"
    excision_code = Dataset()
    excision_code.CodeValue = "65801008"  # 65801008/P1-03000 (CT/RT)
    excision_code.CodingSchemeDesignator = "SCT"  # Snowmed CT defined code
    excision_code.CodeMeaning = "Excision"
    excision.ConceptNameCodeSequence = [excision_name]
    excision.ConceptCodeSequence = [excision_code]

    step_1 = Dataset()
    step_1.SpecimenPreparationStepContentItemSequence = [
        specimen_id,
        spec_id_issuer,
        collection,
        taken,
        excision,
    ]

    dissection = Dataset()
    dissection.ValueType = "CODE"
    dissection_name = Dataset()
    dissection_name.CodeValue = "111704"
    dissection_name.CodingSchemeDesignator = "DCM"
    dissection_name.CodeMeaning = "Sampling Method"
    dissection_code = Dataset()
    dissection_code.CodeValue = "111726"
    dissection_code.CodingSchemeDesignator = "DCM"  # DICOM defined code
    dissection_code.CodeMeaning = "Dissection with entire specimen submission"
    dissection.ConceptNameCodeSequence = [dissection_name]
    dissection.ConceptCodeSequence = [dissection_code]

    parent_specimen = Dataset()
    parent_specimen.ValueType = "TEXT"
    parent_specimen.TextValue = "D18-6003 A-1"
    parent_specimen_name = Dataset()
    parent_specimen_name.CodeValue = "111705"
    parent_specimen_name.CodingSchemeDesignator = "DCM"
    parent_specimen_name.CodeMeaning = "Parent Specimen Identifier"
    parent_specimen.ConceptNameCodeSequence = [parent_specimen_name]

    parent_specimen_type = Dataset()
    parent_specimen_type.ValueType = "CODE"
    parent_specimen_type_name = Dataset()
    parent_specimen_type_name.CodeValue = "111707"
    parent_specimen_type_name.CodingSchemeDesignator = "DCM"
    parent_specimen_type_name.CodeMeaning = "Parent Specimen Type"
    parent_specimen_type_code = Dataset()
    parent_specimen_type_code.CodeValue = "119376003"  # 119376003/G-8300 (CT/RT)
    parent_specimen_type_code.CodingSchemeDesignator = "SCT"
    parent_specimen_type_code.CodeMeaning = "Tissue Specimen"
    parent_specimen_type.ConceptNameCodeSequence = [parent_specimen_type_name]
    parent_specimen_type.ConceptCodeSequence = [parent_specimen_type_code]

    step_2 = Dataset()
    step_2.SpecimenPreparationStepContentItemSequence = [  # 5 substeps
        specimen_id,
        spec_id_issuer,
        dissection,
        parent_specimen,
        parent_specimen_type,
    ]

    formalin_fixing = Dataset()
    formalin_fixing.ValueType = "CODE"
    formalin_fixing_name = Dataset()
    formalin_fixing_name.CodeValue = "111715"
    formalin_fixing_name.CodingSchemeDesignator = "DCM"
    formalin_fixing_name.CodeMeaning = "Specimen Fixative"
    formalin_fixing_code = Dataset()
    formalin_fixing_code.CodeValue = "434162003"  # 434162003/C-2141C (CT/RT)
    formalin_fixing_code.CodingSchemeDesignator = "SCT"
    formalin_fixing_code.CodeMeaning = "Neutral Buffered Formalin"
    formalin_fixing.ConceptNameCodeSequence = [formalin_fixing_name]
    formalin_fixing.ConceptCodeSequence = [formalin_fixing_code]

    step_3 = Dataset()
    step_3.SpecimenPreparationStepContentItemSequence = [  # 3 Steps
        specimen_id,
        spec_id_issuer,
        formalin_fixing,
    ]

    parafin_embedding = Dataset()
    parafin_embedding.ValueType = "CODE"
    parafin_embedding_name = Dataset()
    parafin_embedding_name.CodeValue = "430863003"  # 430863003/F-6221A (CT/RT)
    parafin_embedding_name.CodingSchemeDesignator = "SCT"
    parafin_embedding_name.CodeMeaning = "Embedding Medium"
    parafin_embedding_code = Dataset()
    parafin_embedding_code.CodeValue = "F-616D8"
    parafin_embedding_code.CodingSchemeDesignator = "SCT"
    parafin_embedding_code.CodeMeaning = "Paraffin Wax"
    parafin_embedding.ConceptNameCodeSequence = [parafin_embedding_name]
    parafin_embedding.ConceptCodeSequence = [parafin_embedding_code]

    step_3 = Dataset()
    step_3.SpecimenPreparationStepContentItemSequence = [
        specimen_id,
        spec_id_issuer,
        parafin_embedding,
    ]

    staining = Dataset()
    staining.ValueType = "CODE"
    staining_name = Dataset()
    staining_name.CodeValue = "111701"
    staining_name.CodingSchemeDesignator = "SCT"
    staining_name.CodeMeaning = "Processing Type"
    staining_code = Dataset()
    staining_code.CodeValue = "127790008"  # 127790008/P3-00003 (CT/RT)
    staining_code.CodingSchemeDesignator = "SCT"
    staining_code.CodeMeaning = "Staining"
    staining.ConceptNameCodeSequence = [staining_name]
    staining.ConceptCodeSequence = [staining_code]

    hematoxylin = Dataset()
    hematoxylin.ValueType = "CODE"
    hematoxylin_name = Dataset()
    hematoxylin_name.CodeValue = "424361007"  # 424361007/G-C350 (CT/RT)
    hematoxylin_name.CodingSchemeDesignator = "SCT"
    hematoxylin_name.CodeMeaning = "Using Substance"
    hematoxylin_code = Dataset()
    hematoxylin_code.CodeValue = "12710003"  # 12710003/C-22968 (CT/RT)
    hematoxylin_code.CodingSchemeDesignator = "SCT"
    hematoxylin_code.CodeMeaning = "Hematoxylin Stain"
    hematoxylin.ConceptNameCodeSequence = [hematoxylin_name]
    hematoxylin.ConceptCodeSequence = [hematoxylin_code]

    eosin = Dataset()
    eosin.ValueType = "CODE"
    eosing_name = Dataset()
    eosing_name.CodeValue = "G-C350"
    eosing_name.CodingSchemeDesignator = "SCT"
    eosing_name.CodeMeaning = "Using Substance"
    eosin_code = Dataset()
    eosin_code.CodeValue = "36879007"  # 36879007/C-22919 (CT/RT)
    eosin_code.CodingSchemeDesignator = "SCT"
    eosin_code.CodeMeaning = "Water Soluble Eosin Stain"
    eosin.ConceptNameCodeSequence = [eosing_name]
    eosin.ConceptCodeSequence = [eosin_code]

    step_4 = Dataset()
    step_4.SpecimenPreparationStepContentItemSequence = [
        specimen_id,
        staining,
        hematoxylin,
        eosin,
    ]

    return [
        step_1,
        step_2,
        step_3,
        step_3,
        step_4,
    ]


def brightfield_optical_path_sequence(
    identifier: str = "1",
    icc_profile: Union[bytes, str, Path] = "default",
) -> List[Dataset]:
    """Create a Brightfield Optical Path Sequence.

    Args:
        identifier (str, optional):
            The identifier of the optical path.
        icc_profile (Union[bytes, str, Path], optional):
            The ICC profile of the optical path.

    Returns:
        List[Dataset]:
            The Brightfield Optical Path Sequence.
    """
    optial_path = Dataset()
    optial_path.OpticalPathIdentifier = identifier  # Type 1
    illumination_type_code = CodeRef(
        value="R-102C0",
        designator="SCT",
        meaning="Full Spectrum",
    )
    optial_path.IlluminationTypeCodeSequence = [illumination_type_code]  # Type 1
    if icc_profile == "default":
        icc_profile = importlib.resources.read_binary(
            "wsic.data", "sRGB_v4_ICC_preference.icc"
        )
    if isinstance(icc_profile, (str, Path)):
        optial_path.ICCProfile = Path(icc_profile).read_bytes()
    else:
        optial_path.ICCProfile = icc_profile
    illumination_color_code = CodeRef(
        value="11744",
        designator="DCM",
        meaning="Brightfield Illumination",
    )
    optial_path.IlluminationColorCodeSequence = [illumination_color_code]  # Type 1
    return [optial_path]  # Type 1


def create_vl_wsi_dataset(
    size: Tuple[int, int],
    tile_size: Tuple[int, int],
    photometric_interpretation: Literal["RGB", "YBR_FULL_422"] = "YBR_FULL_422",
) -> Tuple[FileMetaDataset, Dataset]:
    """Create a VL Whole Slide Microscopy Image Storage dataset.

    Currently only supports RGB and YBR_FULL_422 photometric
    interpretations, JPEG transfer syntax, and a single frame.

    Args:
        size:
            The size of the image in pixels as a tuple of (width,
            height).
        tile_size:
            The size of tiles in pixels as a tuple of (width, height).
        photometric_interpretation:
            The photometric interpretation of the image.

    """
    mosaic_size = (
        ceil(size[0] / tile_size[0]),
        ceil(size[1] / tile_size[1]),
    )

    # Create Meta Information Dataset
    meta = FileMetaDataset()
    meta.TransferSyntaxUID = JPEGBaseline
    meta.MediaStorageSOPClassUID = VLWholeSlideMicroscopyImageStorage
    meta.MediaStorageSOPInstanceUID = generate_uid()
    meta.ImplementationClassUID = generate_uid()
    meta.FileMetaInformationVersion = b"\x00\x01"  # See PS3.10
    meta.FileMetaInformationGroupLength = 0  # Will get set later (by dcmwriter)

    # Main dataset
    ds = Dataset()

    # UIDs for study and series instances
    study_instance_uid = generate_uid()
    series_instance_uid = generate_uid()

    # UIDs for SOP and specimen instances
    sop_instance_uid = meta.MediaStorageSOPInstanceUID
    specimen_uid = generate_uid()

    # Patient
    birth_date = datetime.strptime("19000101", "%Y%m%d")
    ds.PatientName = "Test^Patient"  # Type 2
    ds.PatientID = "123456"  # Type 2
    ds.PatientBirthDate = birth_date  # Jan 1st 1970, Type 2
    ds.PatientSex = "O"  # "Other", Type 2

    # General Study
    now = datetime.now()
    ds.StudyDate = now  # Type 2
    ds.StudyTime = now  # Type 2
    ds.StudyID = "123456"  # Type 2
    ds.AccessionNumber = "123456"  # Type 2
    ds.ReferringPhysicianName = "Test^Referring"  # Type 2
    ds.StudyInstanceUID = study_instance_uid  # Type 1

    # General Series
    ds.Modality = "SM"  # "Slide Microscopy", Type 2
    ds.SeriesInstanceUID = series_instance_uid  # Type 1
    ds.SeriesNumber = "1"  # Type 2

    # Whole Slide Microscopy Series

    # Frame of Reference
    ds.FrameOfReferenceUID = generate_uid()  # Type 1
    ds.PositionReferenceIndicator = "SLIDE_CORNER"  # Type 2

    # Multi-frame Dimension
    # (defines a set of dimensions for the multi-frame image)
    # └ Dimension Organization Sequence
    dimension_organization = Dataset()
    dimension_organization.DimensionOrganizationUID = generate_uid()  # Type 1
    ds.DimensionOrganizationSequence = [dimension_organization]

    # General Equipment
    ds.Manufacturer = "Manufacturer"  # Type 2
    ds.PixelPaddingValue = 255  # Type 1C

    # Enhanced General Equipment
    ds.ManufacturerModelName = "Model"  # Type 1
    ds.DeviceSerialNumber = "123456"  # Type 1
    ds.SoftwareVersions = "1.0"  # Type 1

    # General Image
    ds.InstanceNumber = "1"  # Type 2
    ds.ImageType = ["ORIGINAL", "PRIMARY", "VOLUME", "NONE"]  # Type 3

    # Image Pixel
    ds.SamplesPerPixel = 3  # Type 1
    ds.PhotometricInterpretation = photometric_interpretation  # Type 1
    ds.PlanarConfiguration = 0  # Type 1
    ds.Rows = tile_size[1]  # Type 1
    ds.Columns = tile_size[0]  # Type 1
    ds.TotalPixelMatrixRows = size[1]  # Type 1
    ds.TotalPixelMatrixColumns = size[0]  # Type 1
    ds.BitsAllocated = 8  # Type 1
    ds.BitsStored = 8  # Type 1
    ds.HighBit = 7  # Type 1
    ds.PixelRepresentation = 0  # uint, Type 1

    # Acquisition Context, may be empty
    ds.AcquisitionContextSequence = []  # Type 2

    # Multi-Frame Functional Groups
    ds.ContentDate = now  # Type 1
    ds.ContentTime = now  # Type 1
    ds.NumberOfFrames = mosaic_size[0] * mosaic_size[1]  # Type 1
    shared_functional_groups = Dataset()
    shared_functional_groups.ReferencedImageSequence = []
    pixel_measures = Dataset()
    pixel_measures.SliceThickness = 1.0  # Type 1
    pixel_measures.PixelSpacing = [0.0016, 0.0016]  # in mm, Type 1
    shared_functional_groups.PixelMeasuresSequence = [pixel_measures]
    wsi_image_frame_type = Dataset()
    wsi_image_frame_type.FrameType = ["ORIGINAL", "PRIMARY", "VOLUME", "NONE"]  # Type 1
    shared_functional_groups.WholeSlideMicroscopyImageFrameTypeSequence = [
        wsi_image_frame_type,
    ]
    ds.SharedFunctionalGroupsSequence = [shared_functional_groups]  # Type 1

    # Multi-Frame Dimension
    ds.DimensionOrganizationType = "TILED_FULL"  # Type 3

    # Specimen
    ds.ContainerIdentifier = "Container"  # Type 1
    container_id_issuer = Dataset()
    container_id_issuer.LocalNamespaceEntityID = "XYZ Medical Centre"  # Type 1
    ds.IssuerOfTheContainerIdentifierSequence = [container_id_issuer]  # Type 1
    container_type_code = Dataset()
    container_type_code.CodeValue = "430856003"  # 430856003/G-8439 (CT/RT)
    container_type_code.CodingSchemeDesignator = "SCT"  # Snowmed CT defined code
    container_type_code.CodeMeaning = "Tissue Section"
    ds.ContainerTypeCodeSequence = [container_type_code]  # Type 1
    ds.IssuerOfTheContainerIdentifierSequence = []  # Type 1
    specimen_description = Dataset()
    specimen_description.SpecimenIdentifier = "Specimen"  # Type 1
    specimen_description.SpecimenUID = specimen_uid  # Type 1

    # Preparation Sequence
    specimen_description.SpecimenPreparationSequence = (
        ffpe_he_preparation_sequence()
    )  # Type 1
    specimen_description.IssuerOfTheSpecimenIdentifierSequence = []
    ds.SpecimenDescriptionSequence = [specimen_description]  # Type 1

    # Whole Slide Microscopy Image
    ds.AcquisitionDateTime = now  # Type 1
    ds.VolumetricProperties = "VOLUME"  # Type 1
    ds.BurnedInAnnotation = "NO"  # Type 1
    ds.LossyImageCompression = "01"  # Type 1
    ds.LossyImageCompressionMethod = "ISO_10918_1"  # Type 1
    ds.LossyImageCompressionRatio = "30"  # Type 1C
    ds.ImagedVolumeWidth = 1  # mm, Type 1
    ds.ImagedVolumeHeight = 1  # mm, Type 1
    ds.ImagedVolumeDepth = 1  # mm, Type 1
    ds.FocusMethod = "AUTO"  # Type 1
    ds.SpecimenLabelInImage = "NO"  # Type 1
    ds.ExtendedDepthOfField = "NO"  # Type 1
    ds.ImageOrientationSlide = [0, -1, 0, -1, 0, 0]  # Type 1
    ds.TotalPixelMatrixFocalPlanes = 1  # Type 1C
    pixel_matrix_origin = Dataset()
    pixel_matrix_origin.XOffsetInSlideCoordinateSystem = 0  # mm, Type 1
    pixel_matrix_origin.YOffsetInSlideCoordinateSystem = 0  # mm, Type 1
    ds.TotalPixelMatrixOriginSequence = [  # Type 1C
        pixel_matrix_origin,
    ]

    # Optical Path
    ds.OpticalPathSequence = brightfield_optical_path_sequence()  # Type 1
    ds.NumberOfOpticalPaths = 1  # Type 1C

    # SOP Common
    ds.SOPClassUID = VLWholeSlideMicroscopyImageStorage  # Type 1
    ds.SOPInstanceUID = sop_instance_uid  # Type 1

    return meta, ds



"""Tests for the color space enum."""
import pytest

import wsic


def test_condensed():
    for value in wsic.enums.ColorSpace:
        condensed = value.condensed()
        assert " " not in condensed


def test_from_string_fail():
    with pytest.raises(ValueError, match="Unknown color space"):
        wsic.enums.ColorSpace.from_string("unknown")


def test_to_jp2_fail():
    enum = wsic.enums.ColorSpace.MINISWHITE
    with pytest.raises(ValueError, match="no known JP2 equivalent"):
        enum.to_jp2()


def test_from_tiff_fail():
    with pytest.raises(ValueError, match="Unsupported TIFF"):
        wsic.enums.ColorSpace.from_tiff(1234)


def test_from_tiff_aperio():
    """Test special cases for Aperio color space from compression."""
    # Aperio SVS TIFF JPEG2000 YCbCr
    enum = wsic.enums.ColorSpace.from_tiff(..., 33003)
    assert enum is wsic.enums.ColorSpace.YCBCR

    # Aperio SVS TIFF JPEG2000 RGB
    enum = wsic.enums.ColorSpace.from_tiff(..., 33005)
    assert enum is wsic.enums.ColorSpace.RGB


def test_from_tiff_common_cases():
    """Test cases for commmon TIFF colorspaces."""
    enum = wsic.enums.ColorSpace.from_tiff(1)
    assert enum is wsic.enums.ColorSpace.GRAY

    enum = wsic.enums.ColorSpace.from_tiff(2)
    assert enum is wsic.enums.ColorSpace.RGB

    enum = wsic.enums.ColorSpace.from_tiff(6)
    assert enum is wsic.enums.ColorSpace.YCBCR



import multiprocessing
import os
import time
import warnings
from abc import ABC, abstractmethod
from concurrent.futures import ThreadPoolExecutor
from math import ceil, floor
from pathlib import Path
from typing import Iterator, List, Optional, Tuple

import dask.distributed as daskd
import numpy as np
import zarr

from wsic.multiproc import Queue
from wsic.readers import Reader
from wsic.utils import mosaic_shape, tile_slices, wrap_index


def get_tile(
    results_queue: Queue,
    ji: Tuple[int, int],
    tilesize: Tuple[int, int],
    path: Path,
) -> None:
    """Append a tile read from a reader to a multiprocessing queue.

    Args:
        results_queue (Queue):
            A multiprocessing Queue to put tiles on to.
        ji (Tuple[int, int]):
            Index of tile.
        tilesize (Tuple[int, int]):
            Tile size as (width, height).
        path (Path):
            Path to file to read from.

    Returns:
        None
    """
    reader = Reader.from_file(path)
    # Read the tile
    j, i = ji
    index = (
        slice(j * tilesize[1], (j + 1) * tilesize[1]),
        slice(i * tilesize[0], (i + 1) * tilesize[0]),
    )
    # Filter warnings (e.g. from gylmur about reading past the edge)
    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        tile = reader[index]
    results_queue.put((ji, tile))


def get_tile_persistent(
    results_queue: Queue,
    ji_queue: Queue,
    tilesize: Tuple[int, int],
    path: Path,
) -> None:
    """Repeatedly append read tiles to a multiprocessing queue.

    Args:
        results_queue (Queue):
            A multiprocessing Queue to put tiles on to.
        ji_queue (Queue):
            A multiprocessing Queue to get tile indexes from.
        tilesize (Tuple[int, int]):
            Tile size as (width, height).
        path (Path):

    Returns:
        None
    """
    reader = Reader.from_file(path)
    while True:
        ji = ji_queue.get()
        if ji is None:
            break
        # Read the tile
        j, i = ji
        index = (
            slice(j * tilesize[1], (j + 1) * tilesize[1]),
            slice(i * tilesize[0], (i + 1) * tilesize[0]),
        )
        # Filter warnings (e.g. from gylmur about reading past the edge)
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            tile = reader[index]
        results_queue.put((ji, tile))


class TileIterator(ABC):
    """Base class for tile iterators."""

    def __init__(
        self,
        reader: Reader,
        read_tile_size: Tuple[int, int],
        yield_tile_size: Optional[Tuple[int, int]] = None,
        num_workers: int = None,
        intermediate=None,
        verbose: bool = False,
        timeout: float = 10.0,
        match_tile_sizes: bool = True,
    ) -> None:
        self.reader = reader
        self.shape = reader.shape
        self.read_tile_size = read_tile_size
        self.yield_tile_size = yield_tile_size or read_tile_size
        self.intermediate = intermediate
        self.verbose = verbose
        self.timeout = timeout if timeout >= 0 else float("inf")
        self.enqueued = set()
        self.reordering_dict = {}
        self.read_j = 0
        self.read_i = 0
        self.yield_i = 0
        self.yield_j = 0
        self.num_workers = num_workers or os.cpu_count() or 2
        self.read_mosaic_shape = mosaic_shape(
            self.shape,
            self.read_tile_size[::-1],
        )
        self.yield_mosaic_shape = mosaic_shape(
            self.shape,
            self.yield_tile_size[::-1],
        )
        self.remaining_reads = list(np.ndindex(self.read_mosaic_shape))
        self.tile_status = zarr.zeros(self.yield_mosaic_shape, dtype="u1")
        try:
            from tqdm.auto import tqdm

            self.read_pbar = tqdm(
                total=np.prod(self.read_mosaic_shape),
                desc="Reading",
                colour="cyan",
                smoothing=0.01,
            )
        except ImportError:
            self.read_pbar = None

        # Validation and error handling
        read_matches_yield = self.read_tile_size == self.yield_tile_size
        if match_tile_sizes and not read_matches_yield and not self.intermediate:
            raise ValueError(
                f"read_tile_size ({self.read_tile_size})"
                f" != yield_tile_size ({self.yield_tile_size})"
                " and intermediate is not set. An intermediate is"
                " required when the read and yield tile size differ."
            )

    def __len__(self) -> int:
        """Return the number of tiles in the reader."""
        return int(np.prod(self.yield_mosaic_shape))

    def __iter__(self) -> Iterator:
        """Return an iterator for the reader."""
        self.read_j = 0
        self.read_i = 0
        return self

    @property
    def read_index(self) -> Tuple[int, int]:
        """Return the current read index."""
        return self.read_j, self.read_i

    @read_index.setter
    def read_index(self, value: Tuple[int, int]) -> None:
        """Set the current read index."""
        self.read_j, self.read_i = value

    @property
    def yield_index(self) -> Tuple[int, int]:
        """Return the current yield index."""
        return self.yield_j, self.yield_i

    @yield_index.setter
    def yield_index(self, value: Tuple[int, int]) -> None:
        """Set the current yield index."""
        self.yield_j, self.yield_i = value

    def wrap_indexes(self) -> None:
        """Wrap the read and yield indexes."""
        self.read_index, overflow = wrap_index(self.read_index, self.read_mosaic_shape)
        if overflow and self.verbose:
            print("All tiles read.")
        self.yield_index, overflow = wrap_index(
            self.yield_index, self.yield_mosaic_shape
        )
        if overflow:
            if self.verbose:
                print("All tiles yielded.")
            self.close()
            raise StopIteration

    def __next__(self) -> np.ndarray:
        """Return the next tile from the reader."""
        # Ensure a valid read ij index
        self.wrap_indexes()

        # Add tile reads to the queue until the maximum number of
        # workers is reached
        self.fill_queue()

        # Get the next yield tile from the queue
        t0 = time.perf_counter()
        while (time.perf_counter() - t0) < self.timeout:
            # Remove all tiles from the queue into the reordering dict
            self.empty_queue()

            # Remove the next read tile from the reordering dict. May be
            # None if the read tile is not in the reordering dict or if
            # an intermediate is used.
            tile = self.pop_next_read_tile()

            # Return the tile if no intermediate is being used and the
            # tile was in the reordering dict.
            if not self.intermediate and tile is not None:
                return tile

            # Get the next tile from the intermediate. Returns None if
            # intermediate is None or the tile is not in the
            # intermediate.
            tile = self.read_next_from_intermediate()
            if tile is not None:
                return tile

            # Ensure the queue is kept full
            self.fill_queue()

            # Sleep and try again
            time.sleep(0.1)
        warnings.warn(
            "Failed to get next tile after 100 attempts. Dumping debug information.",
            stacklevel=2,
        )
        print(f"Reader Shape {self.reader.shape}")
        print(f"Read Tile Size {self.read_tile_size}")
        print(f"Yield Tile Size {self.yield_tile_size}")
        print(f"Read Mosaic Shape {self.read_mosaic_shape}")
        print(f"Yield Mosaic Shape {self.yield_mosaic_shape}")
        print(f"Read Index {self.read_index}")
        print(f"Yield Index {self.yield_index}")
        print(f"Remaining Reads (:10) {self.remaining_reads[:10]}")
        print(f"Enqueued {self.enqueued}")
        print(f"Reordering Dict (keys) {self.reordering_dict.keys()}")
        if hasattr(self, "queue"):
            print(f"Queue Size {len(self.queue)}")
        intermediate_read_slices = tile_slices(
            index=(self.yield_j, self.yield_i),
            shape=self.yield_tile_size[::-1],
        )
        print(f"Intermediate Read slices {intermediate_read_slices}")
        # Terminate the read processes
        self.close()
        raise IOError(f"Tile read timed out at index {self.yield_index}")

    def read_next_from_intermediate(self) -> Optional[np.ndarray]:
        """Read the next tile from the intermediate file."""
        if self.intermediate is None:
            return None
        intermediate_read_slices = tile_slices(
            index=(self.yield_j, self.yield_i),
            shape=self.yield_tile_size[::-1],
        )
        status = self.tile_status[self.yield_index]
        if np.all(status == 1):  # Intermediate has all data for the tile
            self.tile_status[self.yield_index] = 2
            self.yield_i += 1
            return self.intermediate[intermediate_read_slices]
        return None

    @abstractmethod
    def empty_queue(self) -> None:
        """Remove all tiles from the queue into the reordering dict."""

    @abstractmethod
    def fill_queue(self) -> None:
        """Add tile reads to the queue until the max number of workers is reached."""

    def update_read_pbar(self) -> None:
        """Update the read progress bar."""
        if self.read_pbar is not None:
            self.read_pbar.update()

    def pop_next_read_tile(self) -> Optional[np.ndarray]:
        """Remove the next tile from the reordering dict.

        Returns:
            Optional[np.ndarray]:
                The next tile from the reordering dict if available.
                If an intermediate is being used, or the tile is not
                in the reordering dict, this will be None.
        """
        read_ji = (self.read_j, self.read_i)
        if read_ji in self.reordering_dict:
            self.enqueued.remove(read_ji)
            tile = self.reordering_dict.pop(read_ji)

            # If no intermediate is required, return the tile
            if not self.intermediate:
                if tile is None:
                    raise IOError(f"Tile {read_ji} is None")
                self.read_i += 1
                self.update_read_pbar()
                self.yield_i += 1
                return tile

            # Otherwise, write the tile to the intermediate
            intermediate_write_index = tile_slices(
                index=read_ji,
                shape=self.read_tile_size,
            )
            self.intermediate[intermediate_write_index] = tile
            tile_status_index = tuple(
                slice(max(0, floor(x.start / r)), ceil(x.stop / r))
                for x, r in zip(intermediate_write_index, self.yield_tile_size)
            )
            self.tile_status[tile_status_index] = 1
            self.read_i += 1
            self.update_read_pbar()
        return None

    def close(self):  # noqa: B027
        """Safely end any dependants (threads, processes, and files)."""
        pass  # noqa

    def __del__(self):
        self.close()


class MultiProcessTileIterator(TileIterator):
    """An iterator which returns tiles generated by a reader.

    This is a fancy iterator that uses a multiprocress queue to
    accelerate the reading of tiles. It can also use an intermediate
    file to allow for reading and writing with different tile sizes.

    Args:
        reader (Reader):
            Reader for image.
        read_tile_size (Tuple[int, int]):
            Tile size to read from reader.
        yield_tile_size (Optional[Tuple[int, int]]):
            Tile size to yield. If None, yield_tile_size = read_tile_size.
        num_workers (int):
            Number of workers to use.
        intermediate (Path):
            Intermediate reader/writer to use. Must support random
            access reads and writes.
        verbose (bool):
            Verbose output.

    Yields:
        np.ndarray:
            A tile from the reader.
    """

    def __init__(
        self,
        reader: Reader,
        read_tile_size: Tuple[int, int],
        yield_tile_size: Optional[Tuple[int, int]] = None,
        num_workers: int = None,
        intermediate=None,
        verbose: bool = False,
        timeout: float = 10.0,
        match_tile_sizes: bool = True,
    ) -> None:
        super().__init__(
            reader=reader,
            read_tile_size=read_tile_size,
            yield_tile_size=yield_tile_size,
            num_workers=num_workers,
            intermediate=intermediate,
            verbose=verbose,
            timeout=timeout,
            match_tile_sizes=match_tile_sizes,
        )
        self.processes = {}
        self.queue = Queue()

    def empty_queue(self) -> None:
        """Remove all tiles from the queue into the reordering dict."""
        while not self.queue.empty():
            ji, tile = self.queue.get()
            self.reordering_dict[ji] = tile
            self.processes.pop(ji).join()

    def fill_queue(self) -> None:
        """Add tile reads to the queue until the max number of workers is reached."""
        while len(self.enqueued) < self.num_workers and len(self.remaining_reads) > 0:
            next_ji = self.remaining_reads.pop(0)
            process = multiprocessing.Process(
                target=get_tile,
                args=(
                    self.queue,
                    next_ji,
                    self.read_tile_size,
                    self.reader.path,
                ),
                daemon=True,
            )
            process.start()
            self.processes[next_ji] = process
            self.enqueued.add(next_ji)

    def close(self):
        """Safely end any dependants (threads, processes, and files).

        Close progress bars and join child processes. Terminate children
        if they fail to join after one second.
        """
        if self.read_pbar is not None:
            self.read_pbar.close()
        # Join processes in parallel threads
        if self.processes:
            with ThreadPoolExecutor(len(self.processes)) as executor:
                executor.map(lambda p: p.join(1), self.processes.values(), timeout=2)
            # Terminate any child processes if still alive
            for process in self.processes.values():
                if process.is_alive():
                    process.terminate()


class PersistentMultiProcessTileIterator(TileIterator):
    """An iterator which returns tiles generated by a reader.

    This is a fancy iterator that uses a multiprocress queue to
    accelerate the reading of tiles. It can also use an intermediate
    file to allow for reading and writing with different tile sizes.

    Args:
        reader (Reader):
            Reader for image.
        read_tile_size (Tuple[int, int]):
            Tile size to read from reader.
        yield_tile_size (Optional[Tuple[int, int]]):
            Tile size to yield. If None, yield_tile_size = read_tile_size.
        num_workers (int):
            Number of workers to use.
        intermediate (Path):
            Intermediate reader/writer to use. Must support random
            access reads and writes.
        verbose (bool):
            Verbose output.

    Yields:
        np.ndarray:
            A tile from the reader.
    """

    def __init__(
        self,
        reader: Reader,
        read_tile_size: Tuple[int, int],
        yield_tile_size: Optional[Tuple[int, int]] = None,
        num_workers: int = None,
        intermediate=None,
        verbose: bool = False,
        timeout: float = 10.0,
        match_tile_sizes: bool = True,
    ) -> None:
        super().__init__(
            reader=reader,
            read_tile_size=read_tile_size,
            yield_tile_size=yield_tile_size,
            num_workers=num_workers,
            intermediate=intermediate,
            verbose=verbose,
            timeout=timeout,
            match_tile_sizes=match_tile_sizes,
        )
        self.processes = {}
        self.results_queue = Queue()
        self.index_queue = Queue()

        for i in range(self.num_workers):
            process = multiprocessing.Process(
                target=get_tile_persistent,
                args=(
                    self.results_queue,
                    self.index_queue,
                    self.read_tile_size,
                    self.reader.path,
                ),
                daemon=True,
            )
            process.start()
            self.processes[i] = process

    def empty_queue(self) -> None:
        """Remove all tiles from the queue into the reordering dict."""
        while not self.results_queue.empty():
            ji, tile = self.results_queue.get()
            self.reordering_dict[ji] = tile

    def fill_queue(self) -> None:
        """Add tile reads to the queue until the max number of workers is reached."""
        while len(self.enqueued) < self.num_workers and len(self.remaining_reads) > 0:
            next_ji = self.remaining_reads.pop(0)
            self.index_queue.put(next_ji)
            self.enqueued.add(next_ji)

    def close(self):
        """Safely end any dependants (threads, processes, and files).

        Close progress bars and join child processes. Terminate children
        if they fail to join after one second.
        """
        if self.read_pbar is not None:
            self.read_pbar.close()
        # Join processes in parallel threads
        if self.processes:
            with ThreadPoolExecutor(len(self.processes)) as executor:
                executor.map(lambda p: p.join(1), self.processes.values(), timeout=2)
            # Terminate any child processes if still alive
            for process in self.processes.values():
                if process.is_alive():
                    process.terminate()


class DaskTileIterator(TileIterator):
    """An iterator which returns tiles from a reader which uses dask.

    This iterator is used to read tiles from a reader which uses a dask
    array. It is used to read tiles from a reader in a way which is
    compatible with the TileIterator class.

    This works by dispatching dask futures to read tiles from the reader
    and then waiting for the next futures to complete. This is done in a
    way which ensures that the tiles are returned in the correct order
    and that only a limited number of tiles are in memory at any one
    time (i.e. a limited number of futures are active).

    Reads may be performed in parallel by setting the num_workers.
    Reads may be at a multiple of the yield_tile_size. For some backends
    e.g. a JP2 file this may be faster than reading at the
    yield_tile_size.

    Args:
        reader (Reader):
            The reader with an xarray dataset to read tiles from.
        read_tile_size (Tuple[int, int]):
            The size of the tiles to read.
        yield_tile_size (Optional[Tuple[int, int]]):
            Tile size to yield. If None, yield_tile_size = read_tile_size.
        num_workers (int):
            The number of workers to use for reading tiles.

    Yields:
        np.ndarray:
            A tile from the reader.
    """

    def __init__(
        self,
        reader: Reader,
        read_tile_size: Tuple[int, int],
        yield_tile_size: Optional[Tuple[int, int]] = None,
        num_workers: int = None,
        intermediate=None,
        verbose: bool = False,
        timeout: float = 10.0,
        match_tile_sizes: bool = True,
    ):
        super().__init__(
            reader=reader,
            read_tile_size=read_tile_size,
            yield_tile_size=yield_tile_size,
            num_workers=num_workers,
            timeout=timeout,
            intermediate=intermediate,
            verbose=verbose,
            match_tile_sizes=match_tile_sizes,
        )
        self.array = self.reader._dataset["0"]
        try:
            self.client = daskd.get_client()
        except ValueError:
            self.client = daskd.Client()
        self.client.scatter(self.array)
        self.futures: List[Tuple[Tuple[int, int], daskd.Future]] = []

    def fill_queue(self) -> None:
        """Enqueue futures to read tiles."""
        while len(self.futures) < self.num_workers:
            try:
                next_ji = self.remaining_reads.pop(0)
            except IndexError:
                break
            slices = tile_slices(
                index=next_ji,
                shape=self.read_tile_size,
            )
            future = self.client.submit(
                self.array.__getitem__,
                slices,
            )
            self.futures.append((next_ji, future))
            self.enqueued.add(next_ji)

    def empty_queue(self) -> None:
        """Remove all tiles from the queue into the reordering dict."""
        while self.futures:
            ji, future = self.futures.pop(0)
            tile = future.result().to_numpy()
            self.reordering_dict[ji] = tile
